#include <iostream>
#include <string>
#include <fstream>
#include <sstream>
#include <tr1/unordered_map>
#include <cctype>
#include <set>
#include <omp.h>
#include <algorithm>
#include <functional>
#include <cctype>
#include <locale>
#include <vector>
#include <string>


using namespace::std::tr1;
using namespace::std;

string inputFile="";
string outputFile="";

int kmer = 17;
int minL = 18;
int maxL = 40;
int minE = 1;
int minAL = 100;
int maxAL = 800;
int range = 0;
int threads = 4;
bool isSaveTemp = true;
bool isKeepReverseSeq = true;

static inline std::string &ltrim(std::string &s);
static inline std::string &rtrim(std::string &s);
static inline std::string &trim(std::string &s);
bool strPrepare(string &str);
string revcomp(string str);
int str2int(const string &str);
string int2str (int n);
vector<string> split(const string& src, string delimiter);
string replace(string &src, string srcStr, string desStr);
unordered_map<string,int> readInputFile(string inFileName,int minL, int maxL);
unordered_map<string,int> filterInputSet(unordered_map<string,int> &inMap, int kmer);
void saveFilterData(unordered_map<string,int> &inMap,int kmer);
void assembleDistinctKmer(unordered_map<string,int> &inMap,int kmer);
unordered_map<string,unordered_map<string,int> > matchPairSeek(unordered_map<string,int> &inMap, int kmer);
string assemblyReads(unordered_map<string,unordered_map<string,int> > &matchBinMap,
                     unordered_map<string,int> &inMap,string seq, int kmer, int minE, int minAL,int maxAL);
string localAssemble(string seq,unordered_map<string,int> &needAssembly,
                     unordered_map<string,int> &inMap, int kmer,int minE,int minAL,int maxAL);
string lookOPT(unordered_map<string,int> &candidateMap, int contigSize);
unordered_map<string,unordered_map<string,int> > createBinMap(unordered_map<string,int> &inMap, int kmer);
bool seekMatchP(unordered_map<string,unordered_map<string,int> > &binMap, string seq, int kmer);
unordered_map<string,int> seekMatchB(string seq,unordered_map<string,unordered_map<string,int> > &binMap, int kmer);
bool argParsing(int argc, char* argv[]);
void printUsage(char* argv[]);



static inline std::string &ltrim(std::string &s)
{
    s.erase(s.begin(), std::find_if(s.begin(), s.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
    return s;
}

static inline std::string &rtrim(std::string &s)
{
    s.erase(std::find_if(s.rbegin(), s.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), s.end());
    return s;
}

static inline std::string &trim(std::string &s)
{
    return ltrim(rtrim(s));
}

bool strPrepare(string &str)
{
    bool result = true;
    trim(str);
    int len = str.size();
    if(len<minL || len>maxL)
    {
        result = false;
    }
    else
    {
        for(int i=0; i<len; i++)
        {
            str[i] = toupper(str[i]);
            if(str[i]!='A' && str[i]!='T' && str[i]!='C' && str[i]!='G')
            {
                result = false;
                break;
            }
        }
    }
    return result;
}


string revcomp(string str)
{
    int len = str.size();
    for(int i=0; i<len; i++)
    {
        switch(str[i])
        {
        case 'A':
            str[i] = 'T';
            break;
        case 'T':
            str[i] = 'A';
            break;
        case 'C':
            str[i] = 'G';
            break;
        case 'G':
            str[i] = 'C';
            break;
        }
    }
    char temp ;
    for(int i=0,j=len-1; i<j; i++,j--)
    {
        temp = str[i];
        str[i] = str[j];
        str[j] = temp;
    }
    return str;
}

int str2int(const string &str)
{
    stringstream ss(str);
    int num;
    if((ss>>num).fail())
    {
        return -1;
    }
    return num;
}

string int2str (int n)
{
    stringstream newstr;
    newstr<<n;
    return newstr.str();
}

vector<string> split(const string& src, string delimiter)
{
    vector<string> v;
    if( src.size()==0 || delimiter.size()==0)
    {
        return v;
    }

    long deli_len = delimiter.size();
    long index = string::npos,last_search_position = 0;
    while((index=src.find(delimiter,last_search_position))!=src.npos )
    {
        if(index!=last_search_position)
        {
            v.push_back( src.substr(last_search_position, index-last_search_position));
        }
        last_search_position = index + deli_len;
    }

    string last_one = src.substr(last_search_position);
    if(last_one.size()!=0)
    {
        v.push_back(last_one);
    }

    return v;
}

string replace(string &src, string srcStr, string desStr)
{
    if(srcStr.size()==0)
        return src;

    long index = string::npos,last_search_position = 0;
    while((index=src.find(srcStr,last_search_position))!=string::npos)
    {
        src.replace(index,srcStr.size(),desStr);
        last_search_position = index + desStr.size();
    }

    return src;
}



int main(int argc, char* argv[])
{

    if(!argParsing(argc,argv))
    {
        return -1;
    }

    unordered_map<string,int> inMap = readInputFile(inputFile,minL,maxL);

    omp_set_num_threads(threads);

    if(inMap.size()!=0)
    {
        for(int testKmer=kmer; testKmer<=(kmer+range); testKmer++)
        {
            filterInputSet(inMap, testKmer);
            saveFilterData(inMap, testKmer);
            assembleDistinctKmer(inMap,testKmer);
        }
    }

    return 1;
}

unordered_map<string,int> readInputFile(string inFileName,int minL, int maxL)
{
    unordered_map<string,int> inMap;

    ifstream inFile(inFileName.c_str());
    if(!inFile)
    {
        cout << "can't open the input file!";
    }
    else
    {
        cout << "start reading " << inputFile <<endl;

        string str;
        long count = 0;

        while(getline(inFile,str))
        {
            if(!strPrepare(str)) continue;
            inMap[str]++;
            inMap[revcomp(str)]++;
            if( 0 == (++count)%10000 )
            {
                cout << "**************count:" << count << "************" << endl;
            }
        }

        cout << "finish read file, total reads:" << count << endl;
    }

    return inMap;
}




unordered_map<string,int> filterInputSet(unordered_map<string,int> &inMap, int kmer)
{
    unordered_map<string,int>::iterator inMapIt ;
    unordered_map<string,unordered_map<string,int> > binMap ;
    int i=0, loop = 0;
    string ltag ,rtag;
    volatile bool finishTag = true;

    cout << "begin filter input data, the item size is:" << inMap.size() << endl;

    while(!inMap.empty())
    {

        cout << "****************loop:" << loop++ << "**************" <<endl;
        cout << "****************inMap size:" << inMap.size() << "*****************" << endl;

        binMap = createBinMap(inMap,kmer);


        vector<string> addVec;
        for(inMapIt = inMap.begin(); inMapIt!=inMap.end(); inMapIt++)
        {
            addVec.push_back(inMapIt->first);
        }

        finishTag = true;

        #pragma omp parallel private(i,ltag,rtag)
        {
            #pragma omp for schedule(dynamic,100) nowait
            for(i=0; i<addVec.size(); i++)
            {
                ltag = addVec[i];
                rtag = revcomp(ltag);
                if(seekMatchP(binMap,ltag,kmer) || seekMatchP(binMap,rtag,kmer))
                {
                    if(finishTag)
                        finishTag = false;
                    if(inMap.count(ltag))
                        inMap[ltag] = 0;
                    if(inMap.count(rtag))
                        inMap[rtag] = 0;
                }
            }

        }


        if(finishTag) break;

        inMapIt = inMap.begin();
        while(inMapIt!=inMap.end())
        {
            if(inMapIt->second == 0)
            {
                inMapIt = inMap.erase(inMapIt);
            }
            else
            {
                inMapIt++;
            }
        }




    }

    cout << "finish filter input data, the item size is:" << inMap.size() << endl;
    return inMap;
}

void saveFilterData(unordered_map<string,int> &inMap,int kmer)
{
    if(isSaveTemp)
    {
        ofstream tempFile((inputFile+"."+int2str(kmer)).c_str());
        for(unordered_map<string,int>::iterator inMapIt= inMap.begin(); inMapIt!=inMap.end(); inMapIt++)
        {
            string tag = inMapIt->first;
            int num = inMapIt->second;
            tempFile << ">" << tag << "_" << num << endl;
            tempFile << tag << endl;
        }
    }

}


void assembleDistinctKmer(unordered_map<string,int> &inMap,int kmer)
{
    unordered_map<string,unordered_map<string,int> > matchBinMap = matchPairSeek(inMap,kmer);

    cout << "start assembling reads using k-mer of " << kmer << endl;

    int assNum=0;
    ofstream outFile((outputFile+"."+int2str(kmer)).c_str());

    for(unordered_map<string,int>::iterator inMapIt=inMap.begin(); inMapIt!=inMap.end(); inMapIt++)
    {
        string read = inMapIt->first;
        if(read.size()<=kmer)
            continue;
        if((++assNum)%100==0)
            cout << "." ;


        string viroidSeq = assemblyReads(matchBinMap,inMap,read,kmer,minE,minAL,maxAL);

        if(viroidSeq.size()>=minAL && viroidSeq.size()<=maxAL)
        {
            outFile << ">Contig-" << assNum << endl;
            outFile << viroidSeq << endl;
        }
    }

    outFile.close();
    cout << endl;
}


unordered_map<string,unordered_map<string,int> > matchPairSeek(unordered_map<string,int> &inMap, int kmer)
{

    unordered_map<string,unordered_map<string,int> >  binMap = createBinMap(inMap,kmer);
    unordered_map<string,unordered_map<string,int> >  matchBinMap;

    long count = 0;

    cout << "Pair set search using k-mer of " << kmer << endl;

    for(unordered_map<string,int>::iterator inMapIt = inMap.begin(); inMapIt!=inMap.end(); inMapIt++)
    {
        string read = inMapIt->first;
        unordered_map<string,int> pairMap = seekMatchB(read, binMap,kmer);
        for(unordered_map<string,int>::iterator pairMapIt=pairMap.begin(); pairMapIt!=pairMap.end(); pairMapIt++)
        {
            string pair = pairMapIt->first;
            int position = pair.find(",");
            string aread = pair.substr(0,position);
            string bread = pair.substr(position+1);
            matchBinMap[aread][bread]++;
            if((++count)%100000 == 0)  cout << "." ;
        }
    }
    cout << endl;
    return matchBinMap;
}


string assemblyReads(unordered_map<string,unordered_map<string,int> > &matchBinMap,
                     unordered_map<string,int> &inMap,string seq, int kmer, int minE, int minAL,int maxAL)
{
    unordered_map<string,int> BBshape, needAssembly;
    int xControl=1, pControl=0;

    BBshape[seq]++;

    while(xControl>0)
    {
        unordered_map<string,int> hhain;
        int swyyy = 0;

        for(unordered_map<string,int>::iterator BBshapeIt=BBshape.begin(); BBshapeIt!=BBshape.end(); BBshapeIt++)
        {
            string xName = BBshapeIt->first;

            if(!matchBinMap.count(xName))
                continue;

            unordered_map<string,int>& kkk = matchBinMap[xName];
            for(unordered_map<string,int>::iterator kkkIt=kkk.begin(); kkkIt!=kkk.end(); kkkIt++)
            {
                string pass = kkkIt->first;
                if(matchBinMap.count(xName) && matchBinMap[xName].count(pass))
                {
                    swyyy++;
                    pControl++;
                    hhain[pass]++;
                    needAssembly[pass]++;
                    matchBinMap[xName].erase(pass);
                    if(!isKeepReverseSeq){
                        string rxName = revcomp(xName);
                        string rPass = revcomp(pass);
                        if(matchBinMap.count(rPass))
                            matchBinMap[rPass].erase(rxName);
                    }
                }
            }
        }

        BBshape.clear();
        BBshape = hhain;
        hhain.clear();

        if(swyyy==0)
            xControl=0;
    }

    if(pControl >= 1)
    {
        string longFseq = localAssemble(seq,needAssembly,inMap,kmer,minE,minAL,maxAL);
        return longFseq;
    }
    else
    {
        return "";
    }
}

string localAssemble(string seq,unordered_map<string,int> &needAssembly,
                     unordered_map<string,int> &inMap, int kmer,int minE,int minAL,int maxAL)
{

    int zkmer = kmer + 1;
    unordered_map<string,int> rcount;
    unordered_map<string,unordered_map<string,int> > bbbin;

    for(unordered_map<string,int>::iterator it=needAssembly.begin(); it!=needAssembly.end(); it++)
    {
        string xys = it->first;
        int num = inMap[xys];
        for(int i=0; i<=xys.size(); i++)
        {
            string str1 = xys.substr(i,zkmer);
            string str2 = xys.substr(i,kmer);
            if(str1.size()==zkmer)
            {
                bbbin[str2][str1] += num;
                rcount[str1] += num;
            }
        }
    }



    unordered_map<string,int> candidateAssembleContig;
    unordered_map<string,int> BBshape;
    BBshape[seq] = inMap[seq];
    int xControl = 1;

    while(xControl>0)
    {
        unordered_map<string,int> hhain;
        int swyyy = 0;

        for(unordered_map<string,int>::iterator it=BBshape.begin(); it!=BBshape.end(); it++)
        {
            string xys = it->first;

            if(xys.size()>maxAL)
                continue;

            int pValue = BBshape[xys];
            int startP = xys.size()-kmer;
            string startSeq = xys.substr(startP,kmer);

            if(!bbbin.count(startSeq))
            {
                if(xys.size()<minAL)
                    continue;
                candidateAssembleContig[xys]=pValue;
            }
            else
            {
                unordered_map<string,int>& kkk = bbbin[startSeq];
                for(unordered_map<string,int>::iterator kkkIt=kkk.begin(); kkkIt!=kkk.end(); kkkIt++)
                {
                    string pass = kkkIt->first;
                    if(pass.size()==zkmer)
                    {
                        swyyy++;
                        string newch = pass.substr(kmer,1);
                        int num = rcount[pass];
                        if(num < minE)
                        {
                            continue;
                        }
                        num = pValue;
                        string aabb = xys + newch;
                        hhain[aabb] = num;
                        bbbin[startSeq].erase(pass);
                    }
                }
            }
        }

        int ttas = hhain.size();
        if(ttas>0)
        {
            BBshape.clear();
            BBshape = hhain;
        }

        if(ttas==0 || swyyy==0)
        {
            xControl = 0;
        }
    }

    int contigSize = 0;
    for(unordered_map<string,int>::iterator it=BBshape.begin(); it!=BBshape.end(); it++)
    {
        string yread = it->first;
        candidateAssembleContig[yread] = BBshape[yread];
        contigSize = yread.size();
    }

    string lookSeq = lookOPT(candidateAssembleContig,contigSize);

    return lookSeq;
}


string lookOPT(unordered_map<string,int> &candidateMap, int contigSize)
{

    int qlength = contigSize-25;
    string optContig="",tagName="";

    vector<string> keyVec;
    for(unordered_map<string,int>::iterator candidateMapIt=candidateMap.begin(); candidateMapIt!=candidateMap.end(); candidateMapIt++)
    {
        keyVec.push_back(candidateMapIt->first);
    }

    sort(keyVec.begin(),keyVec.end());

    for(vector<string>::reverse_iterator keyVecIt=keyVec.rbegin(); keyVecIt!=keyVec.rend(); keyVecIt++)
    {
        string yread = (*keyVecIt);

        if(yread.size() >= qlength)
        {
            optContig = yread;
            int ttt = yread.size()-40;
            if(ttt<0)
            {
                ttt = yread.size() + ttt;
            }

            if(optContig=="")
            {
                continue;
            }

            string xseq = optContig.substr(0,40);
            string yseq = optContig.substr(ttt,40);

            for(int i=6; i<=38; i++)
            {
                for(int j=0; j<=xseq.size(); j++)
                {
                    string atag = xseq.substr(j,i);

                    if(atag.size()!=i)
                    {
                        continue;
                    }

                    if(yseq.find(atag,0) != string::npos)
                    {
                        tagName = atag;
                    }
                }
            }

            if(tagName!="") break;
        }
    }

    string finalSeq = "";

    if(tagName!="")
    {
        optContig = replace(optContig,tagName," "+tagName+" ");

        vector<string> seqArray = split(optContig," ");

        int start=0, end=0;

        for(int i=0; i<seqArray.size(); i++)
        {
            if(seqArray[i].substr(0,tagName.size())==tagName)
            {
                start = i;
                break;
            }
        }

        for(int i=seqArray.size()-1; i>=0; i--)
        {
            if(seqArray[i].substr(0,tagName.size())==tagName)
            {
                end = i;
                break;
            }
        }

        if(start>=0 && end>=start)
        {
            for(int i=start; i<end; i++)
            {
                finalSeq = finalSeq + seqArray[i];
            }
        }

        return finalSeq;
    }
}




unordered_map<string,unordered_map<string,int> > createBinMap(unordered_map<string,int> &inMap, int kmer)
{
    unordered_map<string,unordered_map<string,int> > binMap ;
    unordered_map<string,int>::iterator inMapIt = inMap.begin();

    while(inMapIt != inMap.end())
    {
        string second_key = inMapIt->first;
        string first_key = second_key.substr(0,kmer);
        if(!binMap.count(first_key))
        {
            binMap[first_key][second_key]++;
        }

        inMapIt++;
    }

    return binMap;
}



bool seekMatchP(unordered_map<string,unordered_map<string,int> > &binMap, string seq, int kmer)
{
    int match = true;
    int size = seq.size();

    for(int index = size-kmer; index>=0; index--)
    {
        string str1 = seq.substr(index,kmer);
        string str2 = seq.substr(index,size);

        if(!binMap.count(str1))
        {
            continue;
        }
        else
        {
            unordered_map<string,int>& secondMap = binMap[str1];
            unordered_map<string,int>::iterator secondMapIt = secondMap.begin();
            while(secondMapIt != secondMap.end())
            {
                string str = secondMapIt->first;
                if(str.substr(0,str2.size())==str2 && str.size()>str2.size())
                {
                    match = false;
                    goto BACK;
                }
                secondMapIt++;
            }
        }
    }

BACK:
    return match;
}




unordered_map<string,int> seekMatchB(string seq,
                                     unordered_map<string,unordered_map<string,int> > &binMap, int kmer)
{

    unordered_map<string,int> pairMap;
    int size = seq.size();

    for(int index=0; index<=size-kmer; index++)
    {
        string str1 = seq.substr(index,kmer);
        string str2 = seq.substr(index,size);

        if(str1.size()==kmer && str2.size()>=kmer)
        {
            if(!binMap.count(str1)) continue;

            unordered_map<string,int> tempMap = binMap[str1];

            for(unordered_map<string,int>::iterator tempMapIt= tempMap.begin(); tempMapIt!=tempMap.end(); tempMapIt++)
            {
                string pass = tempMapIt->first;
                if(pass.substr(0,str2.size())==str2 && pass.size()>str2.size())
                {
                    pairMap[(seq+","+pass)]++;
                }

            }
        }
    }
    return pairMap;
}





bool argParsing(int argc, char* argv[])
{
    string argStr[argc];
    for(int i=0; i<argc; i++)
    {
        argStr[i] = argv[i];
    }

    if(argc<3)
    {
        printUsage(argv);
        return false;
    }

    for(int i=1; i<argc; i++)
    {
        if(i<argc-1)
        {
            if(argStr[i] == "-f")
            {
                inputFile = argStr[++i];
            }
            else if(argStr[i] == "-o")
            {
                outputFile = argStr[++i];
            }
            else if(argStr[i] == "-k")
            {
                kmer = str2int(argStr[++i]);
                if(kmer == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-s")
            {
                minL = str2int(argStr[++i]);
                if(minL == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-l")
            {
                maxL = str2int(argStr[++i]);
                if(maxL == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-n")
            {
                minE = str2int(argStr[++i]);
                if(minE == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-x")
            {
                range = str2int(argStr[++i]);
                if(minE == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-D")
            {
                minAL = str2int(argStr[++i]);
                if(minE == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-U")
            {
                maxAL = str2int(argStr[++i]);
                if(minE == -1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-p")
            {
                threads = str2int(argStr[++i]);
                if(threads < 1)
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-t")
            {
                string cmd = argStr[++i].substr(0,1);
                if(cmd == "y" || cmd == "Y")
                {
                    isSaveTemp = true;
                }
                else if(cmd == "n" || cmd == "N")
                {
                    isSaveTemp = false;
                }
                else
                {
                    printUsage(argv);
                    return false;
                }
            }
            else if(argStr[i] == "-r")
            {
                string cmd = argStr[++i].substr(0,1);
                if(cmd == "y" || cmd == "Y")
                {
                    isKeepReverseSeq = true;
                }
                else if(cmd == "n" || cmd == "N")
                {
                    isKeepReverseSeq = false;
                }
                else
                {
                    printUsage(argv);
                    return false;
                }

            }
            else
            {
                printUsage(argv);
                return false;
            }
        }
    }

    if(inputFile == "")
    {
        printUsage(argv);
        return false;
    }

    if(outputFile=="")
    {
        outputFile = inputFile + ".viroid";
    }

    return true;
}




void printUsage(char* argv[])
{
    cout << "Usage:" << argv[0]  << endl;
    cout << "-f <file.fasta> input reads" << endl;
    cout << "-o <output_file> sequences of viroid candidates (default: input_file.contigs)" << endl;
    cout << "-k <number> initial assembling k-mer (default = 17)" << endl;
    cout << "-s <number> minimal length of reads for assembling (default = 18)" << endl;
    cout << "-l <number> maximal length of reads for assembling (default = 40)" <<endl;
    cout << "-n <number> the mininal number of reads for extension (default = 1)"<<endl;
    cout << "-x <number> the range of test kmers (i.e., 17+x) (default: +0)"<<endl;
    cout << "-D <number> the mininal length of candidate sequence (default = 100)"<<endl;
    cout << "-U <number> the maximal length of candidate sequence (default = 800)"<<endl;
    cout << "-p <number> the number of threads for paralleling (default = 4)" << endl;
    cout << "-t <y or n> whether save the kmer remainder file (default = y)" << endl;
    cout << "-r <y or n> whether keep the reverse sequence when assemble (default = y)" << endl;
}
